/*
 * webrtc demuxer
 * Copyright (c) 2021, Tencent. Fei Wei<weifei@tencent.com>
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <pthread.h>

#include "libavutil/avstring.h"
#include "libavutil/opt.h"
#include "libavutil/time.h"
#include "libavformat/avformat.h"
#include "libavformat/url.h"

#include "leb_connection_api.h"

#if LIBAVFORMAT_VERSION_MAJOR > 58
#include "libavformat/internal.h"
#else
#define ffstream(avstream)   (avstream)
#endif

/**
 * Guide for webrtc_demuxer
 * 1. depends on libLebConnection SDK from https://github.com/tencentyun/libLebConnectionSDK/
 * 2. ffmpeg integration example, refer to https://docs.qq.com/doc/DWExGWnZacnZWR2JW
 * 3. ijkplayer integration example, refer to https://mp.weixin.qq.com/s/f3ct29ydzAjdJ1fIdOmHmQ
 * 4. parameters in the AVOption table can be set or read from ffplay through option interface
 * 5. offer_ufrag_pwd reported by OnStatsCallback() should be saved and provided for log tracing
 * 6. for Android, must use System.loadLibrary() to load LibLebConnection_so.so explicitly
 */

#define DEBUG 0
#define MAX_ABR_NUM 3

typedef struct PacketQueue {
    AVPacketList *first_pkt, *last_pkt;
    int is_started;
    int nb_packets;
    int abort_request;
    pthread_mutex_t mutex;
    pthread_cond_t condition;
} PacketQueue;

typedef struct WEBRTCContext {
    const AVClass *class;
    AVFormatContext *avctx;
    int open_timeout;
    int read_timeout;

    LebConnectionHandle *handle;
    int video_stream_index_in;
    int video_stream_index_out;
    int audio_stream_index_in;
    int audio_stream_index_out;
    LebVideoCodecType video_codec;
    LebAudioCodecType audio_codec;
    int enable_aac;
    int receive_video;
    int receive_audio;

    // video info
    char *video_extradata;
    int video_extradata_size;
    // audio info
    int sample_rate;
    int num_channels;
    char *audio_extradata;
    int audio_extradata_size;

    AVPacket video_pkt;
    AVPacket audio_pkt;
    PacketQueue queue;
    int max_queue_size;
    int  log_offset;
    int  error_code;

    char *server_address;
    int minisdp_server_port;
    int min_delay_ms;
    int allowed_avsync_diff;
    int max_jitter_delay_ms;

    // 0:关闭abr, 1:abr本地控制, 2:abr服务端控制(要和后台域名配置配合)
    int abr_mode;
    // 从高到低，不同abr清晰度等级，最多3级：high，medium，low，对应不同转码模板，需要后台配置
    // high可以为“origin”， 表示原始流，不转码
    // 最典型应用就是低码率开播，然后根据网络和播放状态调整到合适码率或原始流
    char *video_definitions[MAX_ABR_NUM];
    char *current_video_definition;
    char *requested_video_definition;
    int  requested_definition_index;
    int  max_definition_index;

    char *client_ip;
    char *signal_server_ip;
    char *data_server_ip;

    int dump_file;
    FILE *vfp;
    FILE *afp;
    int64_t video_pts;
    int64_t audio_pts;
} WEBRTCContext;

static int packet_queue_init(PacketQueue *q, void *logctx)
{
    int ret;

    memset(q, 0, sizeof(PacketQueue));
    ret = pthread_mutex_init(&q->mutex, NULL);
    if (ret != 0) {
        ret = AVERROR(ret);
        av_log(logctx, AV_LOG_ERROR, "pthread_mutex_init failed : %s\n", av_err2str(ret));
        return ret;
    }
    ret = pthread_cond_init(&q->condition, NULL);
    if (ret != 0) {
        pthread_mutex_destroy(&q->mutex);
        ret = AVERROR(ret);
        av_log(logctx, AV_LOG_FATAL, "pthread_cond_init failed : %s\n", av_err2str(ret));
        return ret;
    }

    return 0;
}

static void packet_queue_destroy(PacketQueue *q)
{
    AVPacketList *pkt, *nextpkt;

    for (pkt = q->first_pkt; pkt; pkt = nextpkt) {
        nextpkt = pkt->next;
        av_packet_unref(&pkt->pkt);
        av_freep(&pkt);
    }
    q->last_pkt = NULL;
    q->first_pkt = NULL;
    q->nb_packets = 0;
    pthread_mutex_destroy(&q->mutex);
    pthread_cond_destroy(&q->condition);
}

static void packet_queue_abort(PacketQueue *q)
{
    pthread_mutex_lock(&q->mutex);
    q->abort_request = 1;
    pthread_cond_signal(&q->condition);
    pthread_mutex_unlock(&q->mutex);
}

static int packet_queue_wait_start(PacketQueue *q, WEBRTCContext *s, int64_t timeout) {
    int started;
    int loop = (timeout >= 0) ? FFMAX(timeout / 100000, 1) : -1;

    for (int i = 0; loop > 0 && i < loop; i++) {
        if (ff_check_interrupt(&s->avctx->interrupt_callback))
            return AVERROR_EXIT;

        pthread_mutex_lock(&q->mutex);
        if (s->error_code) {
             pthread_mutex_unlock(&q->mutex);
             return s->error_code;
        }
        if (!q->is_started) {
            int64_t t = av_gettime() + 100000;
            struct timespec tv = { .tv_sec  =  t / 1000000,
                    .tv_nsec = (t % 1000000) * 1000 };
            pthread_cond_timedwait(&q->condition, &q->mutex, &tv);
        }
        started = q->is_started;
        pthread_mutex_unlock(&q->mutex);
        if (started)
            return 0;
    }

    return AVERROR(ETIMEDOUT);
}

static int packet_queue_put(PacketQueue *q, AVPacket *pkt, WEBRTCContext *s)
{
    AVPacketList *pkt1;

    pthread_mutex_lock(&q->mutex);
    if (q->abort_request) {
        av_packet_unref(pkt);
        pthread_mutex_unlock(&q->mutex);
        return -1;
    }

    if (s->max_queue_size > 0 && q->nb_packets >= s->max_queue_size) {
        av_log(s->avctx, AV_LOG_INFO, "packet queue overflow, %d/%d, drop old packet\n", q->nb_packets, s->max_queue_size);
        pkt1 = q->first_pkt;
        q->first_pkt = pkt1->next;
        av_packet_unref(&pkt1->pkt);
        av_free(pkt1);
        q->nb_packets--;
    }
    
    pkt1 = av_malloc(sizeof(AVPacketList));
    pkt1->pkt = *pkt;
    pkt1->next = NULL;

    if (!q->last_pkt)
        q->first_pkt = pkt1;
    else
        q->last_pkt->next = pkt1;
    q->last_pkt = pkt1;

    q->is_started = 1;
    q->nb_packets++;
    pthread_cond_signal(&q->condition);

    pthread_mutex_unlock(&q->mutex);
    return 0;
}

static int packet_queue_get(PacketQueue *q, WEBRTCContext *s, AVPacket *pkt, int64_t timeout)
{
    AVPacketList *pkt1;
    int ret = 0;
    int loop = (timeout >= 0) ? FFMAX(timeout / 100000, 1) : -1;

    for (int i = 0; loop > 0 && i < loop; i++) {
        if (ff_check_interrupt(&s->avctx->interrupt_callback)) {
            packet_queue_abort(q);
            ret = AVERROR_EXIT;;
            break;
        }

        pthread_mutex_lock(&q->mutex);
        if (s->error_code) {
             pthread_mutex_unlock(&q->mutex);
             return s->error_code;
        }
        pkt1 = q->first_pkt;
        if (pkt1) {
            q->first_pkt = pkt1->next;
            if (!q->first_pkt)
                q->last_pkt = NULL;
            q->nb_packets--;
            *pkt = pkt1->pkt;
            av_free(pkt1);
            ret = 1;
            pthread_mutex_unlock(&q->mutex);
            break;
        } else {
            int64_t t = av_gettime() + 100000;
            struct timespec tv = { .tv_sec  =  t / 1000000,
                                   .tv_nsec = (t % 1000000) * 1000 };
            ret = AVERROR(ETIMEDOUT);
            pthread_cond_timedwait(&q->condition, &q->mutex, &tv);
        }
        pthread_mutex_unlock(&q->mutex);
    }

    return ret;
}

/* keep it the same as avpriv_set_pts_info */
static void set_stream_pts_info(AVStream *s, int pts_wrap_bits,
                         unsigned int pts_num, unsigned int pts_den)
{
    AVRational new_tb;
    if (av_reduce(&new_tb.num, &new_tb.den, pts_num, pts_den, INT_MAX)) {
        if (new_tb.num != pts_num)
            av_log(NULL, AV_LOG_DEBUG,
                   "st:%d removing common factor %d from timebase\n",
                   s->index, pts_num / new_tb.num);
    } else
        av_log(NULL, AV_LOG_WARNING,
               "st:%d has too large timebase, reducing\n", s->index);

    if (new_tb.num <= 0 || new_tb.den <= 0) {
        av_log(NULL, AV_LOG_ERROR,
               "Ignoring attempt to set invalid timebase %d/%d for st:%d\n",
               new_tb.num, new_tb.den,
               s->index);
        return;
    }
    s->time_base     = new_tb;
#if FF_API_LAVF_AVCTX
    s->codec->pkt_timebase = new_tb;
#endif
    s->pts_wrap_bits = pts_wrap_bits;
}

static int to_ff_log_level(WEBRTCContext *s, LebLogLevel level)
{
    int v = FFMIN((int)level + s->log_offset, (int)kError);

    switch(v) {
    case kDebug:
        return AV_LOG_DEBUG;
    case kInfo:
        return AV_LOG_INFO;
    case kWarning:
        return AV_LOG_WARNING;
    case kError:
        return AV_LOG_ERROR;
    default:
        return AV_LOG_TRACE;
    }
}

static LebLogLevel to_leb_log_level(WEBRTCContext *s)
{
    AVFormatContext *h = s->avctx;
    int v = av_log_get_level();
    LebLogLevel lebloglevel;

    if (v <= AV_LOG_ERROR) {
        lebloglevel = kError;
    } else if (v <= AV_LOG_WARNING) {
        lebloglevel = kWarning;
    } else if (v < AV_LOG_VERBOSE) {
        lebloglevel = kInfo;
    } else { // >= AV_LOG_VERBOSE
        lebloglevel = kDebug;
    }
    av_log(h, AV_LOG_INFO, "av_log_level %d, leb_log_level %d\n", v, lebloglevel);
    return lebloglevel;
}

static void OnLogCallback(void* context, const char* tag, LebLogLevel level, const char* msg) {
    WEBRTCContext *s = context;
    AVFormatContext *h = s->avctx;

    av_log(h, to_ff_log_level(s, level), "[lebconnection]%s%s\n", tag, msg);
}

static void OnVideoInfoCallback(void* context, LebVideoInfo info) {
    WEBRTCContext *s = (WEBRTCContext*)context;
    AVFormatContext *h = s->avctx;

    s->video_codec = info.codec_type;
    if (s->video_codec != kH264 && s->video_codec != kH265 && s->video_codec != kAV1) {
        av_log(h, AV_LOG_ERROR,
                "OnVideoInfoCallback, unknown video codec %d\n",
                s->video_codec);
        return;
    }
    av_log(h, AV_LOG_INFO, "OnVideoInfoCallback, video codec %d, width %zu, hight %zu, extradata_size %zu\n",
           info.codec_type, info.width, info.hight, info.extra_size);

    if (info.extra_size > 0 && !s->video_extradata) {
        s->video_extradata = av_malloc(info.extra_size);
        if (!s->video_extradata) {
            s->error_code = AVERROR(ENOMEM);
            return;
        }
        memcpy(s->video_extradata, info.extra_data, info.extra_size);
        s->video_extradata_size = info.extra_size;

        //just a fake video packet for create stream
        if (h->ctx_flags & AVFMTCTX_NOHEADER) {
            AVPacket *pkt = &s->video_pkt;
            av_new_packet(pkt, 0);
            pkt->stream_index = s->video_stream_index_in;
            packet_queue_put(&s->queue, pkt, s);
        }
    }
}

static void OnAudioInfoCallback(void* context, LebAudioInfo info) {
    WEBRTCContext *s = (WEBRTCContext*)context;
    AVFormatContext *h = s->avctx;

    s->audio_codec = info.codec_type;
    if (s->audio_codec != kAac && s->audio_codec != kOpus) {
        av_log(h, AV_LOG_ERROR,
                "OnAudioInfoCallback, unknown audio codec %d\n",
                s->audio_codec);
        return;
    }
    s->sample_rate = info.sample_rate;
    s->num_channels = info.num_channels;
    av_log(h, AV_LOG_INFO, "OnAudioInfoCallback, audio codec %d, sample rate %d, num channels %d, extradata_size %zu\n",
           info.codec_type, info.sample_rate, info.num_channels, info.extra_size);

    if (info.extra_size > 0 && !s->audio_extradata) {
        s->audio_extradata = av_malloc(info.extra_size);
        if (!s->audio_extradata) {
            s->error_code = AVERROR(ENOMEM);
            return;
        }
        memcpy(s->audio_extradata, info.extra_data, info.extra_size);
        s->audio_extradata_size = info.extra_size;

        //just a fake audio packet for create stream
        if (h->ctx_flags & AVFMTCTX_NOHEADER) {
            AVPacket *pkt = &s->audio_pkt;
            av_new_packet(pkt, 0);
            pkt->stream_index = s->audio_stream_index_in;
            packet_queue_put(&s->queue, pkt, s);
        }
    }
}

static void OnVideoDataCallback(void* context, LebEncodedVideoFrame videoframe)
{
    WEBRTCContext *s = (WEBRTCContext*)context;
    AVPacket *pkt = &s->video_pkt;
#if DEBUG
    av_log(s->avctx, AV_LOG_INFO, "OnVideoDataCallback, size:%zu, is_keyframe %d, dts:%lld, pts:%lld, a-v diff:%lld\n",
           videoframe.size, videoframe.is_keyframe, videoframe.dts, videoframe.pts, s->audio_pts - videoframe.pts);
#endif
    av_new_packet(pkt, videoframe.size);
    memcpy(pkt->data, videoframe.data, videoframe.size);
    pkt->stream_index = s->video_stream_index_in;
    pkt->dts = videoframe.dts;
    pkt->pts = videoframe.pts;
    pkt->flags = AV_PKT_FLAG_KEY;
    s->video_pts = videoframe.pts;
    packet_queue_put(&s->queue, pkt, s);
}

static void OnAudioDataCallback(void* context, LebEncodedAudioFrame audioframe)
{
    WEBRTCContext *s = (WEBRTCContext*)context;
    AVPacket *pkt = &s->audio_pkt;
#if DEBUG
    av_log(s->avctx, AV_LOG_INFO, "OnAudioDataCallback, size:%zu, dts:%lld, pts:%lld, a-v diff:%lld\n",
           audioframe.size, audioframe.dts, audioframe.pts, audioframe.pts - s->video_pts);
#endif
    av_new_packet(pkt, audioframe.size);
    memcpy(pkt->data, audioframe.data, audioframe.size);
    pkt->stream_index = s->audio_stream_index_in;
    pkt->dts = audioframe.dts;
    pkt->pts = audioframe.pts;
    s->audio_pts = audioframe.pts;
    packet_queue_put(&s->queue, pkt, s);
}

static void OnMetaDataCallback(void* context, LebMetaData metadata)
{
    WEBRTCContext *s = (WEBRTCContext*)context;
    AVFormatContext *h = s->avctx;
    av_log(h, AV_LOG_INFO, "OnMetaDataCallback %zu\n", metadata.size);
}

static void OnStatsCallback(void* context, LebStats stats)
{
    WEBRTCContext *s = context;
    AVFormatContext *h = s->avctx;
    /**
     * 请用户保存或上报offer_ufrag_pwd信息，能对应每次播放记录，排查线上问题时需提供此信息，用于查询相关日志
     */
    av_log(h, to_ff_log_level(s, kInfo),
      "OnStats "
      "offer_ufrag_pwd: %s, "
      "client_ip: %s, signal_ip: %s, media_ip:%s, dns_cost: %d, "
      "first video packet delay:%d, first audio packet delay: %d, "
      "rtt:%d audio_bytes:%" PRIu64" audio_nack:%u,audio_lost:%d,audio_pkts:%u, "
      "video_bytes:%" PRIu64" video_nack:%u video_lost:%d,video_pkts:%u,"
      "video_frames_received:%u\n",
      stats.offer_ufrag_pwd, stats.client_ip,
      stats.signal_server_ip, stats.data_server_ip, stats.signal_dns_cost_ms,
      stats.first_video_received_cost_ms, stats.first_audio_received_cost_ms,
      stats.rtt_ms, stats.audio_bytes_received, stats.audio_nack_count,
      stats.audio_packets_lost, stats.audio_packets_received,
      stats.video_bytes_received, stats.video_nack_count,
      stats.video_packets_lost, stats.video_packets_received,
      stats.video_frames_received);

    // 保存客户端IP，信令服务器IP和数据服务器IP
    if (!s->client_ip || av_strcasecmp(s->client_ip, stats.client_ip)) {
        av_free(s->client_ip);
        s->client_ip = av_strdup(stats.client_ip);
    }
    if (!s->signal_server_ip || av_strcasecmp(s->signal_server_ip, stats.signal_server_ip)) {
        av_free(s->signal_server_ip);
        s->signal_server_ip = av_strdup(stats.signal_server_ip);
    }
    if (!s->data_server_ip || av_strcasecmp(s->data_server_ip, stats.data_server_ip)) {
        av_free(s->data_server_ip);
        s->data_server_ip = av_strdup(stats.data_server_ip);
    }
}

static void OnErrorCallbck(void* context, LebErrorCode error)
{
    WEBRTCContext *s = (WEBRTCContext*)context;
    AVFormatContext *h = s->avctx;

    pthread_mutex_lock(&s->queue.mutex);
    if (error == kUnknow_error) {
        s->error_code = AVERROR_UNKNOWN;
    } else if (error == kNotFound_error) {
        s->error_code = AVERROR_STREAM_NOT_FOUND;
    } else if (error == kAuth_error) {
        s->error_code = AVERROR(EINVAL);
    } else if (error == kSignal_timeout) {
        s->error_code = AVERROR(ETIMEDOUT);
    } else if (error == kData_timeout) {
        s->error_code = AVERROR(ETIMEDOUT);
    } else if (error == kReceive_bye) {
        s->error_code = AVERROR_EOF;
    } else if (error != kNone) {
        s->error_code = AVERROR(EINVAL);
    } else {
        s->error_code = 0;
    }
    pthread_mutex_unlock(&s->queue.mutex);

    av_log(h, AV_LOG_ERROR, "OnErrorCallbck %d, %s\n", error, av_err2str(s->error_code));
}

static void OnConnectionChangeCallback(void* context, LebNetState state)
{
    WEBRTCContext *s = (WEBRTCContext*)context;
    AVFormatContext *h = s->avctx;

    av_log(h, AV_LOG_INFO, "OnConnectionChangeCallback %d\n", state);
    pthread_mutex_lock(&s->queue.mutex);
    if (state == kConnectFailed || state == kConnectClosed) {
        s->error_code = AVERROR_EOF;
    }
    pthread_mutex_unlock(&s->queue.mutex);
}

static int webrtc_open(AVFormatContext *h, const char *uri)
{
    WEBRTCContext *s = h->priv_data;
    LebConfig config;
    LebLogLevel level = to_leb_log_level(s);//SDK输出日志等级
    char* new_uri = NULL;
    static const LebCallback callback = {
        .onLogInfo = OnLogCallback,
        .onVideoInfo = OnVideoInfoCallback,
        .onAudioInfo = OnAudioInfoCallback,
        .onEncodedVideo = OnVideoDataCallback,
        .onEncodedAudio = OnAudioDataCallback,
        .onMetaData = OnMetaDataCallback,
        .onStatsInfo = OnStatsCallback,
        .onError = OnErrorCallbck,
        .onConnectionChange = OnConnectionChangeCallback,
    };

    av_log(h, AV_LOG_INFO, "webrtc_open %s\n", uri);
    packet_queue_init(&s->queue, h);
    av_init_packet(&s->video_pkt);
    av_init_packet(&s->audio_pkt);

    memset(&config, 0, sizeof(LebConfig));
    s->video_codec = kNoVideo;
    s->audio_codec = kNoAudio;

    if (av_strstart(uri, "http://", NULL) && strstr(uri, ".sdp")) {
        new_uri = av_strireplace(uri, "http://", "webrtc://");
    } else if (av_strstart(uri, "https://", NULL) && strstr(uri, ".sdp")) {
        new_uri = av_strireplace(uri, "https://", "webrtc://");
    } else if (av_strstart(uri, "webrtc://", NULL))  {
        new_uri = av_strdup(uri);
    } else {
        packet_queue_destroy(&s->queue);
        return AVERROR(EINVAL);
    }
    av_log(h, AV_LOG_INFO, "new uri %s\n", new_uri);

    /**
     * 配置参数
     */
    config.stream_url = new_uri;
    // 通用快直播信令域名, 客户按自己的业务需求设置（计费不同）
    // 全球：overseas-webrtc.liveplay.myqcloud.com(默认)
    // 国内：webrtc.liveplay.myqcloud.com
    // 海外：oswebrtc.liveplay.myqcloud.com
    config.signal_address = s->server_address;
    config.enable_aac = s->enable_aac;
    config.enable_flex_fec = 1;
    config.receive_audio = s->receive_audio;
    config.receive_video = s->receive_video;
    config.stats_period_ms = 5000;
    config.enable_audio_plc = 1;
    config.enable_0rtt = 0;
    config.min_delay_ms = s->min_delay_ms;
    config.allowed_avsync_diff = s->allowed_avsync_diff;
    config.max_jitter_delay_ms = s->max_jitter_delay_ms;

    //init for abr
    config.abr_mode = s->abr_mode;
    if (config.abr_mode > 0) {
        s->max_definition_index = -1;
        for (int i = 0; i < MAX_ABR_NUM ; i++) {
            if (s->video_definitions[i]) {
                config.abr_transcode_names[i] = s->video_definitions[i];
                if (s->current_video_definition
                    && !av_strcasecmp(s->current_video_definition, s->video_definitions[i])) {
                    config.start_transcode_name = s->current_video_definition;
                    s->requested_definition_index = i;
                    s->requested_video_definition = s->video_definitions[s->requested_definition_index];
                }
                s->max_definition_index = i;
            } else {
                break;
            }
        }
        if (s->max_definition_index > 0 && !config.start_transcode_name) {
            // 默认以最低码率为起始模板
            config.start_transcode_name = s->video_definitions[s->max_definition_index];
            s->requested_definition_index = s->max_definition_index;
            s->requested_video_definition = s->video_definitions[s->requested_definition_index];
        }
        av_log(h, AV_LOG_INFO, "init abr definiton %s, index %d\n", s->requested_video_definition, s->requested_definition_index);
    }

    /**
     * 1. 创建LebConnecton, 并注册回调函数
     */
    s->handle = OpenLebConnection((void*)s, level);
    RegisterLebCallback(s->handle, &callback);

    /**
     * 2. 开始连接，内部走信令流程，直接建联拉流；
     */
    StartLebConnection(s->handle, config);

    av_free(new_uri);
    av_log(h, AV_LOG_INFO, "webrtc_open exit\n");
    return 0;
}

static int webrtc_close(AVFormatContext *h)
{
    av_log(h, AV_LOG_INFO, "webrtc_close\n");
    WEBRTCContext *s = h->priv_data;

    if (s->vfp) {
        fclose(s->vfp);
        s->vfp = NULL;
    }
    if (s->afp) {
        fclose(s->afp);
        s->afp = NULL;
    }

    packet_queue_abort(&s->queue);

    /**
     * 3. 停止和关闭LebConnection
     */
    StopLebConnection(s->handle);
    CloseLebConnection(s->handle);
    s->handle = NULL;

    packet_queue_destroy(&s->queue);

    av_free(s->video_extradata);
    av_free(s->audio_extradata);
    s->video_extradata = NULL;
    s->audio_extradata = NULL;

    av_log(h, AV_LOG_INFO, "webrtc_close exit\n");
    return 0;
}

static AVStream *create_stream(AVFormatContext *s, int codec_type)
{
    WEBRTCContext *h   = s->priv_data;
    av_log(s, AV_LOG_INFO, "create_stream, codec_type %d\n", codec_type);

    AVStream *st = avformat_new_stream(s, NULL);
    if (!st)
        return NULL;
    st->codecpar->codec_type = codec_type;

    if (s->nb_streams >= h->receive_audio + h->receive_video)
        s->ctx_flags &= ~AVFMTCTX_NOHEADER;

    if (codec_type == AVMEDIA_TYPE_VIDEO) {
        if (h->video_codec == kH264) {
            st->codecpar->codec_id = AV_CODEC_ID_H264;
            if (h->dump_file) {
                h->vfp = fopen("video.h264", "wb");
            }
        } else if (h->video_codec == kH265) {
            st->codecpar->codec_id = AV_CODEC_ID_H265;
            if (h->dump_file) {
                h->vfp = fopen("video.h265", "wb");
            }
        } else if (h->video_codec == kAV1) {
            st->codecpar->codec_id = AV_CODEC_ID_AV1;
            if (h->dump_file) {
                h->vfp = fopen("video.av1", "wb");
            }
        }
        ffstream(st)->need_parsing = AVSTREAM_PARSE_NONE;
        h->video_stream_index_out = st->index;

        if (h->video_extradata_size > 0) {
            st->codecpar->extradata = av_malloc(h->video_extradata_size + AV_INPUT_BUFFER_PADDING_SIZE);
            if (st->codecpar->extradata != NULL) {
                memset(st->codecpar->extradata + h->video_extradata_size, 0, AV_INPUT_BUFFER_PADDING_SIZE);
                memcpy(st->codecpar->extradata, h->video_extradata, h->video_extradata_size);
                st->codecpar->extradata_size = h->video_extradata_size;
            } else {
                st->codecpar->extradata_size = 0;
            }
        }
    }

    if (codec_type == AVMEDIA_TYPE_AUDIO) {
        if (h->audio_codec == kAac) {
            st->codecpar->codec_id = AV_CODEC_ID_AAC;
            if (h->dump_file) {
                h->afp = fopen("audio.aac", "wb");
            }
        } else if (h->audio_codec == kOpus) {
            st->codecpar->codec_id = AV_CODEC_ID_OPUS;
            if (h->dump_file) {
                h->afp = fopen("audio.opus", "wb");
            }
        }
        st->codecpar->sample_rate = h->sample_rate;
        st->codecpar->channels = h->num_channels;
        ffstream(st)->need_parsing = AVSTREAM_PARSE_NONE;
        h->audio_stream_index_out = st->index;

        if (h->audio_extradata_size > 0) {
            st->codecpar->extradata = av_malloc(h->audio_extradata_size + AV_INPUT_BUFFER_PADDING_SIZE);
            if (st->codecpar->extradata != NULL) {
                memset(st->codecpar->extradata + h->audio_extradata_size, 0, AV_INPUT_BUFFER_PADDING_SIZE);
                memcpy(st->codecpar->extradata, h->audio_extradata, h->audio_extradata_size);
                st->codecpar->extradata_size = h->audio_extradata_size;
            } else {
                st->codecpar->extradata_size = 0;
            }
        }
    }
    set_stream_pts_info(st, 64, 1, 1000);/* 64 bit pts in ms */
    return st;
}

static int webrtc_probe(const AVProbeData *p)
{
    /**
     * 快直播webrtc stream url形式
     * 1. webrtc://xxxx/xx(.sdp/.flv)
     * 2. http(s)://xxxx/xx.sdp
     */
    if (av_strstart(p->filename, "webrtc://", NULL) ||
        (av_strstart(p->filename, "http://", NULL) && strstr(p->filename, ".sdp")) ||
        (av_strstart(p->filename, "https://", NULL) && strstr(p->filename, ".sdp"))) {
        return AVPROBE_SCORE_MAX;
    }
    return 0;
}

static int webrtc_read_header(AVFormatContext *s)
{
    WEBRTCContext *h = s->priv_data;
#if LIBAVFORMAT_VERSION_MAJOR > 57
    const char *url = s->url;
#else
    const char *url = s->filename;
#endif
    int ret;

    av_log(s, AV_LOG_INFO, "webrtc_read_header, url %s\n", url);
    s->flags |= AVFMT_FLAG_GENPTS;
    s->ctx_flags |= AVFMTCTX_NOHEADER;
    s->fps_probe_size = 0;
    s->max_analyze_duration = FFMAX(s->max_analyze_duration, 5*AV_TIME_BASE);
    s->probesize = FFMAX(s->probesize, 512*1024);
    h->avctx = s;
    h->video_stream_index_in = 0;
    h->audio_stream_index_in = 1;
    h->video_stream_index_out = -1;
    h->audio_stream_index_out = -1;
    ret = webrtc_open(s, url);
    if (ret) {
        av_log(s, AV_LOG_ERROR, "webrtc_read_header: webrtc_open failed, %s\n", av_err2str(ret));
        return ret;
    }

    // 一定时间内收不到数据，超时退出
    ret = packet_queue_wait_start(&h->queue, h, INT64_C(1000) * h->open_timeout);
    if (ret) {
        av_log(s, AV_LOG_ERROR, "webrtc_read_header wait failed, %s\n", av_err2str(ret));
        webrtc_close(s);
        return ret;
    }

    av_log(s, AV_LOG_INFO, "webrtc_read_header exit\n");
    return 0;
}

static int webrtc_read_packet(AVFormatContext *s, AVPacket *pkt)
{
    int ret, i;
    WEBRTCContext *h = s->priv_data;
    AVStream *st;

    // adjust abr tagrt
    if (h->abr_mode == 1) {
        // 端侧控制abr, 播放器可以通过setoption设置current_video_definition来切换码率
        if (h->current_video_definition && h->requested_video_definition
            && av_strcasecmp(h->current_video_definition, h->requested_video_definition)) {
            for (int i = 0; i <= h->max_definition_index; i++) {
                if (!av_strcasecmp(h->current_video_definition, h->video_definitions[i])) {
                    h->requested_definition_index = i;
                    h->requested_video_definition = h->video_definitions[h->requested_definition_index];
                    DoLebCommand(h->handle, "request_abr", (void*)h->requested_video_definition);
                    av_log(h, AV_LOG_INFO, "select abr defintion %s, index %d", h->requested_video_definition, h->requested_definition_index);
                    break;
                }
            }
        }
    }

    do {
        ret = packet_queue_get(&h->queue, h, pkt, INT64_C(1000) * h->read_timeout);
        if (ret < 0)
            break;

        /* now find stream */
        for (i = 0; i < s->nb_streams; i++) {
            st = s->streams[i];
            if (pkt->stream_index == h->video_stream_index_in
                && st->codecpar->codec_type == AVMEDIA_TYPE_VIDEO) {
                break;
            } else if (pkt->stream_index == h->audio_stream_index_in
                       && st->codecpar->codec_type == AVMEDIA_TYPE_AUDIO) {
                break;
            }
        }
        if (i == s->nb_streams) {
            static const enum AVMediaType stream_types[] = {AVMEDIA_TYPE_VIDEO, AVMEDIA_TYPE_AUDIO};
            st = create_stream(s, stream_types[pkt->stream_index]);
            if (!st) {
                av_packet_unref(pkt);
                ret = AVERROR(ENOMEM);
                break;
            }
        }

        if (pkt->size <= 0) {
            // drop fake packet and continue to read packet
            av_packet_unref(pkt);
            ret = 0;
            continue;
        }

        if (pkt->stream_index == h->video_stream_index_in) {
            pkt->stream_index = h->video_stream_index_out;
            if (h->vfp) {
                fwrite(pkt->data, 1, pkt->size, h->vfp);
                fflush(h->vfp);
            }
        } else if (pkt->stream_index == h->audio_stream_index_in) {
            pkt->stream_index = h->audio_stream_index_out;
            if (h->afp) {
                fwrite(pkt->data, 1, pkt->size, h->afp);
                fflush(h->afp);
            }
        } else {
            ret = 0;
        }

        if (!ret) {
            av_log(s, AV_LOG_INFO, "drop pkt with index %d and continue\n",
                   pkt->stream_index);
            av_packet_unref(pkt);
        }
    } while (!ret);

    ret = ret > 0 ? 0 : ret;
    if (ret)
        av_log(s, AV_LOG_WARNING, "webrtc_read_packet, %s\n", av_err2str(ret));
    return ret;
}

static int webrtc_read_close(AVFormatContext *s)
{
    av_log(s, AV_LOG_INFO, "webrtc_read_close\n");
    webrtc_close(s);
    av_log(s, AV_LOG_INFO, "webrtc_read_close exit\n");
    return 0;
}


#define OFFSET(x) offsetof(WEBRTCContext, x)
#define D AV_OPT_FLAG_DECODING_PARAM
#define E AV_OPT_FLAG_ENCODING_PARAM
static const AVOption options[] = {
    { "webrtc_log_offset", "Adjust WebRTC to FFmpeg logging level mapping, negative to reduce verbosity and vice versa", OFFSET(log_offset), AV_OPT_TYPE_INT, { .i64 = 0 }, -4, 3, D },
    { "webrtc_max_queue_size", "The maximum number of packets can hold before dropping, default to -1 for unlimited", OFFSET(max_queue_size), AV_OPT_TYPE_INT, { .i64 = -1 }, -1, INT_MAX, D },
    { "webrtc_dump_file", "Dump video and audio raw data to file", OFFSET(dump_file), AV_OPT_TYPE_BOOL, { .i64 = 0 }, 0, 1, D },
    { "webrtc_server_address", "Webrtc server address", OFFSET(server_address), AV_OPT_TYPE_STRING, { .str = "overseas-webrtc.liveplay.myqcloud.com" }, 0, 0, D },
    { "webrtc_min_delay", "the duration in ms of the data to be sent for the first rollback GOP", OFFSET(min_delay_ms), AV_OPT_TYPE_INT, { .i64 = 1000 }, -1, INT_MAX, D },
    { "webrtc_allowed_avsync_diff", "webrtc allowed avsync diff", OFFSET(allowed_avsync_diff), AV_OPT_TYPE_INT, { .i64 = 100 }, -1, INT_MAX, D },
    { "webrtc_abr_mode", "webrtc abr mode", OFFSET(abr_mode), AV_OPT_TYPE_INT, { .i64 = 0 }, 0, 2, D },
    { "webrtc_high_video_definition", "Webrtc high video definition", OFFSET(video_definitions[0]), AV_OPT_TYPE_STRING, { .str = "origin" }, 0, 0, D },
    { "webrtc_medium_video_definition", "Webrtc medium video definition", OFFSET(video_definitions[1]), AV_OPT_TYPE_STRING, { .str = NULL }, 0, 0, D },
    { "webrtc_low_video_definition", "Webrtc low video definition", OFFSET(video_definitions[2]), AV_OPT_TYPE_STRING, { .str = NULL }, 0, 0, D },
    { "webrtc_current_video_definition", "Webrtc current video definition", OFFSET(current_video_definition), AV_OPT_TYPE_STRING, { .str = NULL }, 0, 0, D },
    { "webrtc_enable_aac", "Webrtc enable aac instead of opus", OFFSET(enable_aac),  AV_OPT_TYPE_INT, { .i64 = 1 }, 0, 1, D },
    { "webrtc_receive_video", "Webrtc receive video", OFFSET(receive_video),  AV_OPT_TYPE_INT, { .i64 = 1 }, 0, 1, D },
    { "webrtc_receive_audio", "Webrtc receive audio", OFFSET(receive_audio),  AV_OPT_TYPE_INT, { .i64 = 1 }, 0, 1, D },
    { "webrtc_open_timeout", "Webrtc open timeout in ms", OFFSET(open_timeout),  AV_OPT_TYPE_INT, { .i64 = 10000 }, 0, INT_MAX, D },
    { "webrtc_read_timeout", "Webrtc read timeout in ms", OFFSET(read_timeout),  AV_OPT_TYPE_INT, { .i64 = 5000 }, 0, INT_MAX, D },
    { "webrtc_max_jitter_delay", "Webrtc max internal jitter delay in ms", OFFSET(max_jitter_delay_ms), AV_OPT_TYPE_INT, { .i64 = 3000 }, -1, INT_MAX, D },
    { "webrtc_client_ip", "Webrtc client ip", OFFSET(client_ip), AV_OPT_TYPE_STRING, { .str = NULL }, 0, 0, AV_OPT_FLAG_EXPORT | AV_OPT_FLAG_READONLY },
    { "webrtc_signal_server_ip", "Webrtc signal server ip", OFFSET(signal_server_ip), AV_OPT_TYPE_STRING, { .str = NULL }, 0, 0, AV_OPT_FLAG_EXPORT | AV_OPT_FLAG_READONLY },
    { "webrtc_data_server_ip", "Webrtc data server ip", OFFSET(data_server_ip), AV_OPT_TYPE_STRING, { .str = NULL }, 0, 0, AV_OPT_FLAG_EXPORT | AV_OPT_FLAG_READONLY },
    { NULL }
};

static const AVClass webrtc_class = {
    .class_name = "webrtc",
    .item_name  = av_default_item_name,
    .option     = options,
    .version    = LIBAVUTIL_VERSION_INT,
};

AVInputFormat ff_webrtc_demuxer = {
    .name           = "webrtc",
    .long_name      = "webrtc demuxer",
    .priv_data_size = sizeof(WEBRTCContext),
    .read_probe     = webrtc_probe,
    .read_header    = webrtc_read_header,
    .read_packet    = webrtc_read_packet,
    .read_close     = webrtc_read_close,
    .extensions      = "webrtc",
    .priv_class     = &webrtc_class,
    .flags          = AVFMT_NOFILE,
};

#ifdef BUILD_AS_PLUGIN
void register_webrtc_demuxer()
{
    av_log(NULL, AV_LOG_INFO, "register_webrtc_demuxer\n");
    av_register_input_format(&ff_webrtc_demuxer);
}
#endif
